@@ -1,32 +1,45 @@
 /*
- *	$Source: /home/nlfm/Working/Frink/RCS/token.c,v $
- *	$Date: 1998/12/04 16:40:45 $
- *	$Revision: 1.2.1.30 $
+ *	$Source: /home/nlfm/Working/CVS/frink/token.c,v $
+ *	$Date: 2001/07/10 15:46:34 $
+ *	$Revision: 2.2 $
  *
  *------------------------------------------------------------------------
  *   AUTHOR:  Lindsay Marshall <lindsay.marshall@newcastle.ac.uk>
  *------------------------------------------------------------------------
- *    Copyright 1994 The University of Newcastle upon Tyne (see COPYRIGHT)
+ *    Copyright 1994-2003 The University of Newcastle upon Tyne (see COPYRIGHT)
  *========================================================================
  *
  */
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
 
 #include <ctype.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "frink.h"
+#include <frink.h>
+#include <blocks.h>
+#include <vars.h>
 
-struct tbuff_s
+#ifdef HAVE_STDLIB_H
+#include <stdlib.h>
+#endif
+
+#ifdef HAVE_MALLOC_H
+#include <malloc.h>
+#endif
+
+extern char *currentfile;
+
+typedef struct tbuff_s
 {
     char *buff;
     int   size;
     int   length;
     int   unit;
-};
-
-typedef struct tbuff_s TBuff;
+} TBuff;
 
 static TBuff *newTB(int size)
 {
@@ -113,7 +126,9 @@ static char *tokenName[] =
     "SCOMMENT",
     "NOSP",
     "SRBRACE",
-    "RPAREN"
+    "RPAREN",
+    "SLBRACE",
+    "PRAGMA"
 };
 
 
@@ -121,77 +136,114 @@ static Token *concToken(char, char, Input *);
 static Token *callToken(Input *);
 static Token *varToken(Input *);
 
-void warn(Token *tp, char *msg)
+Token *newToken(TokenType t)
 {
-    fprintf(stderr, "*** Warning : %s", msg);
-    if (tp != (Token *) 0)
+    Token *tp = malloc(sizeof(Token));
+    tp->type = t;
+    tp->ckind = CNSTWORD;
+    tp->isconc = 0;
+    tp->text = (char *) 0;
+    tp->length = 0;
+    tp->sequence = tp->next = noToken;
+    tp->lineNo = 0;
+    return tp;
+}
+
+void dumpToken(Token *tk, FILE *ops)
+{
+    static char *cval [] =
     {
-	if (tp->lineNo != 0)
-	{
-	    fprintf(stderr, ", line %d", tp->lineNo);
-	}
-	fprintf(stderr, "\nToken Information : %s \"%s\"", tokenName[tp->type],
-	  (tp->text == (char *) 0) ? "(NULL)" : tp->text);
+        "word", "string", "list"
+    };
+    Token *tp;
+    static int depth = -1;
+    if (tk == noToken)
+    {
+	fprintf(ops, "Null Token Pointer\n");
     }
-    fprintf(stderr, "\n");
-    fflush(stderr);
-    if (haltonWarn)
+    else
     {
-	
-	exit(1);
+	depth += 1;
+	fprintf(ops, "{%s ", tokenName[tk->type]);
+	if (tk->type == CONST)
+	{
+	    fprintf(ops, "(%s) ", cval[tk->ckind]);
+	}
+	if (tk->isconc)
+	{
+	    fprintf(ops, "(conc) ");
+	}
+	fprintf(ops, "line %d {", tk->lineNo);
+	if (tk->text != (char *) 0)
+	{
+	    fprintf(ops, "%s", tk->text);
+	}
+	else
+	{
+	    fprintf(ops, "(NULL POINTER)");
+	}
+	fprintf(ops, "} {");
+	tp = tk->sequence;
+	while (tp != noToken)
+	{
+	    dumpToken(tp, ops);
+	    tp = tp->next;
+	}
+	fprintf(ops,"} }");
+	if ((depth -= 1) < 0) { fprintf(ops, "\n"); }
     }
 }
 
 void fail(Token *tp, char *msg)
 {
-    fprintf(stderr, "*** Error : %s", msg);
-    if (tp != (Token *) 0)
+    if (!compout)
     {
+	fprintf(stderr, "***");
+    }
+    if (currentfile != (char *) 0)
+    {
+        fprintf(stderr, " %s", currentfile);
+    }
+    if (!compout)
+    {
+        fprintf(stderr, " Error : %s", msg);
+    }
+    if (tp != noToken)
+    {
+	if (tp->text != (char *) 0)
+	{
+	    fprintf(stderr, "\n    %s\n", tp->text);
+	}
 	if (tp->lineNo != 0)
 	{
-	    fprintf(stderr, ", line %d", tp->lineNo);
+	    fprintf(stderr, (compout ? " (%d) :" : " (line %d)"), tp->lineNo);
 	}
-	fprintf(stderr, "\nToken Information : %s \"%s\"", tokenName[tp->type],
-	  (tp->text == (char *) 0) ? "(NULL)" : tp->text);
+    }
+    if (compout)
+    {
+	fprintf(stderr, " Error : %s", msg);
     }
     fprintf(stderr, "\n");
     exit(1);
 }
 
-void dumpToken(Token *tk)
+extern void makeTB(Token *hd, TBuff *tb);
+ 
+static void makeFTB(Token *hd, TBuff *tb)
 {
-    Token *tp;
-    static int depth = -1;
-    if (tk == 0)
+    if (hd == noToken)
     {
-	fprintf(stderr, "Null Token Pointer\n");
+        return;
     }
-    else
-    {
-	depth += 1;
-	fprintf(stderr, "{%s line %d {", tokenName[tk->type], tk->lineNo);
-	if (tk->text)
-	{
-	    fprintf(stderr, "%s", tk->text);
-	}
-	fprintf(stderr, "} {");
-	tp = tk->sequence;
-	while (tp != (Token *) 0)
-	{
-	    dumpToken(tp);
-	    tp = tp->next;
-	}
-	fprintf(stderr,"} }");
-	if ((depth -= 1) < 0) { fprintf(stderr, "\n"); }
-    }
-}
-
-static void makeTB(Token *hd, TBuff *tb)
-{
-    if (hd == (Token *) 0) { return; }
 
     switch (hd->type)
     {
+    case LBRACE :
+	catTB(tb, "{");
+        break;
+    case RBRACE :
+	catTB(tb, "}");
+        break;
     case VAR :
 	catTB(tb, "$");
 	makeTB(hd->sequence, tb);
@@ -204,7 +256,7 @@ static void makeTB(Token *hd, TBuff *tb)
     case SCOMMENT :
         catTB(tb, ";");
     	catTB(tb, hd->text);
-	catTB(tb, "\n");
+/*	catTB(tb, "\n"); */
 	break;
     case SEMI :
 	catTB(tb, ";");
@@ -215,31 +267,121 @@ static void makeTB(Token *hd, TBuff *tb)
     case CONC :
 	makeTB(hd->sequence, tb);
 	break;
+    case CONST:
+        switch (hd->ckind)
+	{
+	case CNSTWORD:
+	    if (hd->text != (char *) 0) { catTB(tb, hd->text); }
+	    break;
+
+	case CNSTSTRING:
+	    catTB(tb, "\"");
+	    if (hd->text != (char *) 0) { catTB(tb, hd->text); }
+	    catTB(tb, "\"");
+	    break;
+
+	case CNSTLIST:
+	    catTB(tb, "{");
+	    if (hd->text != (char *) 0) { catTB(tb, hd->text); }
+	    catTB(tb, "}");
+	    break;
+	}
+        break;
+
     case STRING :
 	catTB(tb, "\"");
 	makeTB(hd->sequence, tb);
 	catTB(tb, "\"");
 	break;
+
     case SLIST :
 	catTB(tb, "{");
 	makeTB(hd->sequence, tb);
 	catTB(tb, "}");
 	break;
+
     case CALL :
 	catTB(tb, "[");
 	makeTB(hd->sequence, tb);
 	catTB(tb, "]");
 	break;
+
     case LIST :
 	catTB(tb, "{");
 	if (hd->text != (char *) 0) { catTB(tb, hd->text); }
 	catTB(tb, "}");
 	break;
+
     default :
 	if (hd->text != (char *) 0) { catTB(tb, hd->text); }
 	break;
     }
-    makeTB(hd->next, tb);
+}
+
+void makeTB(Token *hd, TBuff *tb)
+{
+    makeFTB(hd, tb);
+    if (hd != noToken) { makeTB(hd->next, tb); }
+}
+
+void warn(Token *tp, char *msg)
+{
+    Blox *blp = (Blox *) lpeek(blocks);
+    TBuff *tbp = newTB(32);
+
+    if (!compout)
+    {
+        fprintf(stderr, "***");
+    }
+    if (currentfile != (char *) 0)
+    {
+        fprintf(stderr, " %s", currentfile);
+    }
+    if (!compout)
+    {
+        fprintf(stderr, " Warning : %s", msg);
+    }
+    if (tp != noToken)
+    {
+        if (compout  && tp->lineNo != 0)
+        {
+            fprintf(stderr, " (%d) :", tp->lineNo);
+        }
+        if (blp != (Blox *) 0 && blp->name != noToken)
+        {
+            fprintf(stderr, " in proc ");
+            makeFTB(blp->name, tbp);
+            fprintf(stderr, "%s", tbp->buff);
+            freeTB(tbp);
+        }
+        if (!compout && tp->lineNo != 0)
+        {
+            fprintf(stderr, " (line %d)", tp->lineNo);
+        }
+    }
+    else if (lineNumber != 0)
+    {
+        fprintf(stderr, (compout ? " (%d) :" : " (line %d)"), lineNumber);
+    }
+    if (compout)
+    {
+        fprintf(stderr, " Warning : %s", msg);
+    }
+    fprintf(stderr, "\n");
+    fflush(stderr);
+    if (haltonWarn)
+    {
+        exit(1);
+    }
+    failed = 1;
+}
+
+static void bwarn(int ln, char *msg)
+{
+    Token *tp = newToken(SP);
+    tp->lineNo = ln;
+    warn(tp, msg);
+    freeToken(tp);
 }
 
 void streamMore(Input *file)
@@ -257,53 +399,91 @@ static int isBlank(char ch)
     return ch == ' ' || ch == '\t';
 }
 
+static char *skipblank(char *cp)
+{
+    while (isBlank(*cp))
+    {
+	cp++;
+    }
+    return cp;
+}
+
 static char reallyGet(Input *file)
 {
     char ch;
     if (file->remaining <= 0) {	streamMore(file); }
     if (file->remaining <= 0) { return '\0'; }
     file->remaining -= 1;
-    if ((ch = *file->position++) == '\n') { file->lineNumber += 1; }
+    if ((ch = *file->position++) == '\n')
+    {
+        file->lineNumber += 1;
+    }
     return ch;
 }
 
-static void push(Input *file, char ch)
+static char push(Input *file, char ch)
 {
     if (ch != '\0')
     {
 	file->back[file->pushed] = ch;
 	file->pushed += 1;
+	if (ch == '\n') { file->lineNumber -= 1; }
     }
+    return ch;
+}
+
+static char getPushed(Input * file)
+{
+    file->pushed -= 1;
+    if (file->back[file->pushed] == '\n')
+    {
+	file->lineNumber += 1;
+    }
+    return file->back[file->pushed];
+}
+
+static char chCGet(Input *file)
+{
+    return (file->pushed) ? getPushed(file) : reallyGet(file);
+}
+
+
+static char peek(Input *file)
+{
+    return push(file, chCGet(file));
 }
 
 static char chGet(Input *file)
 {
     char ch;
-    if (file->pushed)
+    if (file->pushed) { return getPushed(file); }
+    if ((ch = chCGet(file)) == '\\')
     {
-	file->pushed -= 1;
-	return file->back[file->pushed];
-    }
-    else if ((ch = reallyGet(file)) == '\\')
-    {
-	switch (ch = reallyGet(file))
+	switch (ch = chCGet(file))
 	{
 	case '\r':
-	    if ((ch = reallyGet(file)) != '\n' && !isBlank(ch))
+	    if ((ch = chCGet(file)) != '\n' && !isBlank(ch))
 	    {
-	    	push(file,ch);
+	    	(void) push(file,ch);
 		ch = ' ';
 		break;
 	    }
 	case '\n':
-	    while (isBlank(ch = reallyGet(file)))
+	    while (isBlank(ch = chCGet(file)))
 	    {
 	    }
-	    push(file, ch);
+	    (void) push(file, ch);
 	    ch = ' ';
 	    break;
 	default:
-	    push(file, ch);
+	    if (ch == ' ')
+	    {
+	        if (peek(file) == '\n')
+		{
+		    bwarn(file->lineNumber, "\"\\ \" at end of line");
+		}
+	    }
+	    (void) push(file, ch);
 	    ch = '\\';
 	    break;
 	}
@@ -312,30 +492,9 @@ static char chGet(Input *file)
     return ch;
 }
 
-static char chCGet(Input *file)
-{
-    if (file->pushed)
-    {
-	file->pushed -= 1;
-	return file->back[file->pushed];
-    }
-    return reallyGet(file);
-}
-
-Token *newToken(TokenType t)
-{
-    Token *tp = malloc(sizeof(Token));
-    tp->lineNo = 0;
-    tp->type = t;
-    tp->text = (char *) 0;
-    tp->length = 0;
-    tp->sequence = tp->next = (Token *) 0;
-    return tp;
-}
-
 void freeToken(Token *tp)
 {
-    if (tp != (Token *) 0)
+    if (tp != noToken)
     {
 	if (tp->text != (char *) 0 && tp->text != tp->little)
 	{
@@ -347,14 +506,48 @@ void freeToken(Token *tp)
     }
 }
 
+static char *filterString(Token *tp, char *txt)
+{
+    char *cp = malloc(strlen(txt) + 1), *pt;
+    
+    if (tp->type == CONST && tp->ckind != CNSTLIST)
+    {
+        pt = cp;
+	while (*txt)
+	{
+	    if (*txt != '\\')
+	    {
+	        *pt++ = *txt++;
+	    }
+	    else
+	    {
+	        txt++;
+		if (*txt == '\0')
+		{ /* this is a very weird case!!!! */
+		    *pt++ = '\\';
+		}
+		else
+		{
+	            *pt++ = *txt++;
+		}
+	    }
+	}
+	*pt = '\0';
+    }
+    else
+    {
+        strcpy(cp, txt);
+    }
+    return cp;
+}
+
 Input *tokenise(Token *tp, int expr)
 {
     Input *file = (Input *) malloc(sizeof(Input));
-    file->position = tp->text;
     file->remaining = tp->length;
     file->pushed = 0;
-    file->text = tp->text;
     file->stream = NULL;
+    file->position = file->text = filterString(tp, tp->text);
     file->tcall = 0;
     file->texpr = expr;
     file->lineNumber = tp->lineNo;
@@ -362,18 +555,35 @@ Input *tokenise(Token *tp, int expr)
     return file;
 }
 
+Input *ftokenise(char *txt, int length, int lineNo, int expr)
+{
+    Token *tmp = newToken(CONST); /* gash token for filterstring */
+    Input *file = (Input *) malloc(sizeof(Input));
+    file->remaining = length;
+    file->pushed = 0;
+    file->stream = NULL;
+    file->position = file->text = filterString(tmp, txt);
+    file->tcall = 0;
+    file->texpr = expr;
+    file->lineNumber = lineNo;
+    file->lineStart = 1;
+    freeToken(tmp); 
+    return file;
+}
+
 void untokenise(Input *file)
 {
+    free(file->text);
     free(file);
 }
 
-static Token *tokenPop(Token **tp)
+Token *tokenPop(Token **tp)
 {
     Token *sp;
-    if ((sp = *tp) != (Token *) 0)
+    if ((sp = *tp) != noToken)
     {
 	*tp = sp->next;
-	sp->next = (Token *) 0;
+	sp->next = noToken;
     }
     return sp;
 }
@@ -381,14 +591,14 @@ static Token *tokenPop(Token **tp)
 void tokenPush(Token **tp, Token *v)
 {
     Token * ptr;
-    v->next = (Token *) 0;
-    if ((ptr = *tp) == (Token *) 0)
+    v->next = noToken;
+    if ((ptr = *tp) == noToken)
     {
 	*tp = v;
     }
     else
     {
-	while (ptr->next != (Token *) 0)
+	while (ptr->next != noToken)
 	{
 	    ptr = ptr->next;
 	}
@@ -396,16 +606,17 @@ void tokenPush(Token **tp, Token *v)
     }
 }
 
-Token *createToken(TokenType t, char *text, Token *nxt)
+static Token *createToken(TokenType t, char *text, Token *nxt)
 {
     Token *tp = newToken(t);
-    if ((tp->length = strlen(text)) < sizeof(tp->little))
+
+    if ((tp->length = strlen(text)) < (sizeof(tp->little) - 1))
     {
 	tp->text = strcpy(tp->little, text);
     }
     else
     {
-	tp->text = strcpy((char *) malloc(tp->length+1), text);
+	tp->text = newString(text);
     }
     tp->next = nxt;
     return tp;
@@ -426,13 +637,12 @@ static Token *makeToken(TokenType t, TBuff *tb, int ln)
 	free(tb);
     }
     tp->lineNo = ln;
-    
     return tp;
 }
 
 static Token *spaceToken(char ch, Input *file)
 {
-    TBuff *tb = newTB(128);
+     TBuff *tb = newTB(128);
     int ln = file->lineNumber;
     do
     {
@@ -450,29 +660,24 @@ static Token *listToken(Input *file, int term)
     char ch;
     TBuff *tb = newTB(bufSize);
     Token *tp;
-    int bcount = 0, inString = 0, ln = file->lineNumber;
+    int bcount = 0, ln = file->lineNumber;
     
     for (;;)
     {
-	switch (ch = chGet(file))
+      switch (ch = chCGet(file)) /* needs to be chCGet here */
 	{
 	case '\\' :
 	    addTB(tb, ch);
-	    ch = chGet(file);
+	    ch = chCGet(file);  /* ditto - this avoids backslash processing */
 	    break;
-/* I cant remember why this was here and it seems to be broken!!!
-	case '"' :
-	    inString = !inString;
-	    break;
-*/
 	case '{' :
-	    if (!inString) { bcount += 1; };
+	    bcount += 1;
 	    break;
 	case '}' :
-	    if (!inString) { if ((bcount -= 1) < 0) { goto done; } }
+	    if ((bcount -= 1) < 0) { goto done; }
 	    break;
 	case '\0' :
-	    warn((Token *) 0, "Missing }");
+	    bwarn(file->lineNumber, "Missing }");
 	    goto done;
 	}
 	addTB(tb, ch);
@@ -489,20 +694,18 @@ done:
 	    switch (ch)
 	    {
 	    case ')' :
-	        if (file->texpr)
+	        if (!file->texpr)
 		{
-		    break;
+		    fail(tp, "Extra characters after '}'. (Missing '(' ?)");
 		}
-		fail(tp, "Extra characters after '}'. (Missing '(' ?)");
 		break;
 	    case ';' :
 	    case '\0': break;
 	    case ']' :
-		if (file->tcall)
+		if (!file->tcall)
 		{
-		    break;
+		    fail(tp, "Extra characters after '}'. (Missing '[' ?)");
 		}
-		fail(tp, "Extra characters after '}'. (Missing '[' ?)");
 		break;
 	    default :
 		fail(tp, "Extra characters after '}'.");
@@ -523,6 +726,7 @@ Token *stringToken(Input *file, char lst, TokenType res, int term)
     char ch;
     TBuff *tb = newTB(bufSize);
     static char errmsg [] ="Missing X";
+    int variable = (res == ARRAY);
 
     tp->lineNo = file->lineNumber;
     for(;;)
@@ -541,17 +745,18 @@ Token *stringToken(Input *file, char lst, TokenType res, int term)
 	case ';' :  nt = handleSemi(file, lst);  break;
 	case ' ' :
 	case '\t':  nt = spaceToken(ch, file); break;
-	case '[' :  nt = callToken(file); break;
+	case '[' :  nt = callToken(file); variable = 1; break;
 	case '$' :
-	    if ((nt = varToken(file)) == (Token *) 0)
+	    if ((nt = varToken(file)) == noToken)
 	    {
 		addTB(tb, '$');
 		continue;
 	    }
+	    variable = 1;
 	    break;
 	case '\0' :
 	    errmsg[8] = lst;
-	    warn((Token *) 0, errmsg);
+	    bwarn(file->lineNumber, errmsg);
 	    goto done;
 	case '{' : nt = newToken(LBRACE); break;
 	case '}' : nt = newToken(RBRACE); break;
@@ -604,6 +809,13 @@ done:
     }
     
     capTB(tb);
+    if (res != ARRAY && tp->sequence == noToken)
+    {
+	freeToken(tp);
+	tp = makeToken(CONST, tb, file->lineNumber);
+	tp->ckind = CNSTSTRING;
+	return tp;
+    }
     if (tb->length != 0)
     {
 	tokenPush(&tp->sequence, makeToken(SPART, tb, file->lineNumber));
@@ -612,12 +824,62 @@ done:
     {
 	freeTB(tb);
     }
+    if (!variable)
+    {
+	tb = newTB(bufSize);
+	makeTB(tp->sequence, tb);
+	freeToken(tp);
+	tp = makeToken(CONST, tb, file->lineNumber);
+	tp->ckind = CNSTSTRING;
+    }
     return tp;
 }
 
+
+int tokEqual(Token *tp, char *val)
+{
+    if (tp != noToken)
+    {
+        switch (tp->type)
+	{
+	case CONST:
+	case LIST :
+	    return (tp->text != (char *) 0 && strcmp(tp->text, val) == 0);
+	default:
+	    break;
+	}
+    }
+    return 0;
+}
+
+int tokNEqual(Token *tp, char *val, int n)
+{
+    if (tp != noToken)
+    {
+        switch (tp->type)
+	{
+	case CONST:
+	case LIST :
+	    return (tp->text != (char *) 0 && strncmp(tp->text, val, n) == 0);
+	default:
+	    break;
+	}
+    }
+    return 0;
+}
+
+static int isPragma(TBuff *tp)
+{
+    char *cp;
+    
+    cp = skipblank(tp->buff + 1); /* start after # character */
+    return (strncmp(cp, "PRAGMA", 6) == 0 || strncmp(cp, "FRINK", 5) == 0);
+}
+
 static Token *getComment(Input *file, TBuff *tb, char lst)
 {
     char ch;
+    TokenType tt = COMMENT;
 
     for(;;)
     {
@@ -650,7 +912,11 @@ static Token *getComment(Input *file, TBuff *tb, char lst)
 xit:
     push(file, ch);
     capTB(tb);
-    return makeToken(COMMENT, tb, file->lineNumber);
+    if (isPragma(tb))
+    {
+       tt = PRAGMA;
+    }
+    return makeToken(tt, tb, file->lineNumber);
 }
 
 static Token *handleSemi(Input *file, char term)
@@ -666,9 +932,9 @@ static Token *handleSemi(Input *file, char term)
     }
     catTB(tb = newTB(128), "#");
     tp = getComment(file, tb, term);
-    if (term == '\n' && (ch = chCGet(file)) == '\r')
+    if (term == '\n' && chCGet(file) == '\r')
     {
-        ch = chCGet(file);
+        (void) chCGet(file);
     }
     tp->type = SCOMMENT;
     return tp;
@@ -718,8 +984,14 @@ Token *getToken(Input *file)
 	}
     default  : tp = concToken(ch, '\0', file); break;
     }
-    if (tp->lineNo == 0) { tp->lineNo = file->lineNumber; }
-    if (trace) dumpToken(tp);
+    if (tp->lineNo == 0)
+    {
+        tp->lineNo = file->lineNumber;
+    }
+    if (trace)
+    {
+        dumpToken(tp, stderr);
+    }
     return tp;
 }
 
@@ -740,7 +1012,7 @@ static Token *callToken(Input *file)
 	    freeToken(stp);
 	    continue;
 	case ENDF :
-	    warn((Token *) 0, "Missing ]");
+	    bwarn(file->lineNumber, "Missing ]");
 	case ENDC :
 	    freeToken(stp);
 	    file->tcall -= 1;
@@ -760,14 +1032,15 @@ static Token *callToken(Input *file)
  * 1) a sequence of alphanumerics or _
  * 2) a list enclosed in {}
  * 3) a name followed by an array reference
+ * 4) a namespace reference
  * 
  */
 static Token *varToken(Input *file)
 {
     Token *tp = newToken(VAR);
-    int bufSize = 512, leng, sexpr;
+    int bufSize = 512, leng, sexpr, nsc;
     
-    char ch;
+    char ch, ch2;
     TBuff *tb;
 
     sexpr = file->texpr;
@@ -781,24 +1054,65 @@ static Token *varToken(Input *file)
     }
 
     tb = newTB(bufSize);
-    while (isalnum(ch) || ch == '_')	/* alphanumeric name */
+    nsc = 0;
+    while (isalnum(ch) || ch == '_' || ch == ':') /* alphanumeric name */
     {
+        if (ch == ':')
+	{
+	    nsc += 1;
+	}
+	else
+	{
+	    if (doTest(HNSNAME))
+	    {
+	        if (nsc == 1)
+	        {
+	            warn(tp, "Single ':' in variable name!");
+	        }
+		else if (nsc > 2)
+		{
+	            warn(tp, "More than two consecutive ':' characters in variable name!");
+		}
+	    }
+	    nsc = 0;
+	}
 	addTB(tb, ch);
   	ch = chGet(file);
     }
+/* have to back off trailing :s */
+    if (nsc > 0)
+    {
+        push(file, ch);
+	tb->length -= nsc;
+        while (nsc > 1)
+        {
+	    push(file, ':');
+	    nsc -= 1;
+	}
+	ch = ':';
+    }
     capTB(tb);
     if ((leng = tb->length) == 0)
     {
+        if (doTest(HNAMERR))
+        {
+            ch2 = peek(file);
+            if (isalnum(ch2) || ch2 == '_' || ch2 == ':')
+            {
+
+                warn(tp, "Possible variable name error");
+            }
+        }
 	file->texpr = sexpr;
 	push(file, ch);
 	freeToken(tp);
 	freeTB(tb);
-	return (Token *) 0;
+	return noToken;
     }
     tokenPush(&tp->sequence, makeToken(VNAME, tb, file->lineNumber));
     if (ch == '(')
     {
-	if (leng == 0) { fail((Token *) 0, "Variable error"); }
+	if (leng == 0) { fail(noToken, "Variable error"); }
 	tokenPush(&tp->sequence, stringToken(file, ')', ARRAY, 0));
     }
     else
@@ -809,10 +1123,43 @@ static Token *varToken(Input *file)
     return tp;
 }
 
+static void concopt(Token **conc)
+{
+    Token *tp = (*conc)->sequence;
+
+    tp->lineNo = (*conc)->lineNo;
+	
+    if (tp->next == noToken)
+    {
+	switch (tp->type)
+	{
+	case CPART:
+	    tp->type = CONST;
+	    tp->ckind = CNSTWORD;
+	    tp->isconc = 1;
+	    (*conc)->sequence = noToken;
+	    freeToken(*conc);
+	    *conc = tp;
+	    break;
+
+	case CALL:
+	case VAR:
+	    tp->isconc = 1;
+	    (*conc)->sequence = noToken;
+	    freeToken(*conc);
+	    *conc = tp;
+	    break;
+
+	default:
+	    break;
+	}
+    }
+}
+
 static Token *concToken(char fst, char lst, Input *file)
 {
     Token *tp = newToken(CONC), *vp;
-    int bufSize = 512;
+    int bufSize = 512, ln;
     TBuff *tb = newTB(bufSize);
     char ch;
 
@@ -854,7 +1201,7 @@ static Token *concToken(char fst, char lst, Input *file)
 	    ch = chGet(file);
 	    continue;
 	case '$' :
-	    if ((vp = varToken(file)) != (Token *) 0)
+	    if ((vp = varToken(file)) != noToken)
 	    {
 		if (tb->length != 0)
 		{
@@ -883,20 +1230,119 @@ done :
     if (tb->length != 0)
     {
 	capTB(tb);
-	if (tp->sequence == (Token *) 0)
+	if (tp->sequence == noToken)
 	{
+	    ln = tp->lineNo;
 	    freeToken(tp);
-	    return makeToken(CONST, tb, tp->lineNo);
+	    return makeToken(CONST, tb, ln);
 	}
 	tokenPush(&tp->sequence, makeToken(CPART, tb, tp->lineNo));
     }
     else
     {
 	freeTB(tb);
+	
     }
+    concopt(&tp);
     return tp;
 }
 
+static void handlePragma(Token *tp)
+{
+    static struct pragma_s
+    {
+        char		*text;
+	enum Pragmas	value;
+	int		nextmode;
+    } pragmas[] =
+    {
+    	{ "notreached",		NOTREACHED,	1},
+	{ "nocheck",		NOCHECK,	1},
+	{ "noformat",		NOFORMAT,	1},
+	{ "returnok",		RETURNOK,	1},
+	{ "set",		0,		2},
+	{ "unused",		0,		3},
+	{ "array",		0,		4},
+	{ (char *) 0,		0,		1},
+    };
+    struct pragma_s *prp;
+    int label, mode = 1;
+    char *cp;
+    Token *token, *chars = newToken(CONST);
+    Input *pfile;
+
+    pragma = 0;
+    cp = skipblank(tp->text+1);
+    chars->text = cp;
+    chars->length = strlen(cp);
+    pfile=tokenise(chars, 0);
+    token = getToken(pfile);
+    if (tokNEqual(token, "PRAGMA", 6))
+    {
+        label = 1;
+    }
+    else if (tokNEqual(token, "FRINK", 5))
+    {
+        label = 2;
+    }
+    else
+    {
+        warn(tp, "Pragma syntax weirdness");
+    }
+    freeToken(token);
+
+    for(;;)
+    {
+        token = getToken(pfile);
+	switch(token->type)
+	{
+	    case ENDF :
+	    case COMMENT :
+	        goto done;
+
+	    case CONST:
+	    case LIST:
+	        switch (mode)
+		{
+		case 1:
+        	    for (prp = &pragmas[0]; prp->text != (char *) 0; prp++)
+        	    {
+ 	                if (tokEqual(token, prp->text))
+		        {
+		            pragma |= prp->value;
+			    mode = prp->nextmode;
+		            break;
+		        }
+		    }
+		    break;
+
+		case 2: /* mark as set */
+		    setVar(token, VLOCAL, 0);
+		    break;
+
+		case 3: /* mark as used */
+		    useVar(token, VLOCAL, 0);
+		    break;
+
+		case 4: /* mark as an array */
+		    setVar(token, VLOCAL, 1);
+		    break;
+		    
+		}
+	        break;
+
+	    default:
+	        break;
+	}
+	freeToken(token);
+    }
+done:
+    freeToken(token);
+    chars->text = (char *) 0;
+    freeToken(chars);
+    untokenise(pfile);
+}
+
 int handle(Token *line)
 {
     Token *hd;
@@ -904,9 +1350,12 @@ int handle(Token *line)
     extern int tclop(Token*, Token*);
     extern void comment(Token *);
 
-    if (line == (Token *) 0)
+    if (line == noToken)
     {
-	if (!minimise) { blank(); }
+	if (!minimise)
+	{
+	    blank();
+	}
 	return 1;
     }
     hd = tokenPop(&line);
@@ -916,12 +1365,25 @@ int handle(Token *line)
 	freeToken(hd);
 	return 0;
     case CONST :
-	if (tclop(hd, line)) { break; }
+	if (tclop(hd, line))
+	{
+	    break;
+	}
+	/* DROP THROUGH!!!!! */
     case VAR :
     case CALL :
     case CONC :
+        if (isUnreachable())
+	{
+	    warn(hd, "statement is unreachable");
+	    setUnreachable(0);
+	}
 	makeCall(hd, line);
 	break;
+    case PRAGMA :
+        handlePragma(hd);
+	comment(hd);
+	goto exit;
     case COMMENT :
 	comment(hd);
 	break;
@@ -943,6 +1405,8 @@ int handle(Token *line)
 	fail(hd, "Invalid tcl program");
 	/*NOT REACHED*/
     }
+    pragma = 0;
+exit:
     freeToken(hd);
     freeToken(line);
     return 1;
@@ -950,7 +1414,7 @@ int handle(Token *line)
 
 Token *collect(Input *file)
 {
-    Token *line = (Token *) 0, *tp;
+    Token *line = noToken, *tp;
     TBuff *tb;
     for(;;)
     {
@@ -961,17 +1425,17 @@ Token *collect(Input *file)
 	case SEMI:
 	case NL :
 	case ENDF :
-	    if (line == (Token *) 0) { return tp; }
+	    if (line == noToken) { return tp; }
 	    freeToken(tp);
 	    file->lineStart = 1;
 	    return line;
 	case SCOMMENT:
-	    if (line == (Token *) 0) { return tp; }
+	    if (line == noToken) { return tp; }
 	    tokenPush(&line, tp);
 	    file->lineStart = 1;
 	    return line;
 	case CONST :
-	    if (line == (Token *) 0 && tp->text[0] == '#')
+	    if (line == noToken && tp->text[0] == '#')
 	    {
 		tb = newTB(128);
 		catTB(tb, tp->text);
@@ -980,7 +1444,7 @@ Token *collect(Input *file)
 	    }
 	    break;
 	case CONC :
-	    if (line == (Token *) 0 && tp->sequence->type == CPART &&
+	    if (line == noToken && tp->sequence->type == CPART &&
 		tp->sequence->text[0] == '#')
 	    {
 		tb = newTB(128);
@@ -1002,22 +1466,11 @@ Token *collect(Input *file)
 
 static void sconv(Token **conc, Token **line)
 {
-    Token *tp;
-    
-    if (*conc != (Token *) 0)
+    if (*conc != noToken)
     {
-	tp = (*conc)->sequence;
-	tp->lineNo = (*conc)->lineNo;
-	
-	if (tp->type == CPART && tp->next == (Token *) 0)
-	{
-	    tp->type = CONST;
-	    (*conc)->sequence = (Token *) 0;
-	    freeToken(*conc);
-	    *conc = tp;
-	}
+        concopt(conc);
 	tokenPush(line, *conc);
-	*conc = (Token *) 0;
+	*conc = noToken;
     }
 }
 
@@ -1033,7 +1486,7 @@ static Token *listconv(Token **lst)
 	switch ((*lst)->type)
 	{
 	case ENDF :
-	    fail((Token *) 0, "String body too complex");
+	    fail(noToken, "String body too complex");
 	    /*NOT REACHED*/
 	case RBRACE :
 	    tp->lineNo = (*lst)->lineNo;
@@ -1049,8 +1502,8 @@ static Token *listconv(Token **lst)
 
 void sprocess(Token *lst, int nls)
 {
-    Token *line = (Token *) 0, *conc = (Token *) 0;
-    while (lst != (Token *) 0)
+    Token *line = noToken, *conc = noToken;
+    while (lst != noToken)
     {
 	switch (lst->type)
 	{
@@ -1064,14 +1517,14 @@ void sprocess(Token *lst, int nls)
 	case SEMI:
 	case NL :
 	    sconv(&conc, &line);
-	    if (nls || line != (Token *) 0) { handle(line); }
-	    line = (Token *) 0;
+	    if (nls || line != noToken) { handle(line); }
+	    line = noToken;
 	    freeToken(tokenPop(&lst));
 	    continue;
 	case SCOMMENT :
 	    sconv(&conc, &line);
-	    if (nls || line != (Token *) 0) { handle(line); }
-	    line = (Token *) 0;
+	    if (nls || line != noToken) { handle(line); }
+	    line = noToken;
 	    freeToken(tokenPop(&lst));
 	    continue;
 	case SP :
@@ -1081,37 +1534,37 @@ void sprocess(Token *lst, int nls)
 	default :
 	    break;
 	}
-	if (conc == (Token *) 0) { conc = newToken(CONC); }
+	if (conc == noToken) { conc = newToken(CONC); }
 	tokenPush(&conc->sequence, tokenPop(&lst));
     }
     sconv(&conc, &line);
-    if (line != (Token *) 0) { handle(line); }
+    if (line != noToken) { handle(line); }
 }
 
 void lprocess(Token *lst, int nls)
 {
-    Token *line = (Token *) 0;
+    Token *line = noToken;
 
-    while (lst != (Token *) 0)
+    while (lst != noToken)
     {
 	switch (lst->type)
 	{
 	case ENDF :
 	    freeToken(lst);
-	    lst = (Token *) 0;
+	    lst = noToken;
 	    break;
 
 	case SEMI :
 	case NL :
-	    if (nls || line != (Token *) 0) { handle(line); }
-	    line = (Token *) 0;
+	    if (nls || line != noToken) { handle(line); }
+	    line = noToken;
 	case SP :
 	    freeToken(tokenPop(&lst));
 	    break;
 	case SCOMMENT:
 	    tokenPush(&line, tokenPop(&lst));
 	    handle(line);
-	    line = (Token *) 0;
+	    line = noToken;
 	    break;
 	    
 	default :
@@ -1119,12 +1572,12 @@ void lprocess(Token *lst, int nls)
 	    break;
 	}
     }
-    if (line != (Token *) 0) { handle(line); }
+    if (line != noToken) { handle(line); }
 }
 
 Token *accumulate(Input *file, int nl)
 {
-    Token *line = (Token *) 0, *hd;
+    Token *line = noToken, *hd = noToken;
     TokenType last = NL;
     TBuff *tb;
     for(;;)
@@ -1133,7 +1586,7 @@ Token *accumulate(Input *file, int nl)
 	switch (hd->type)
 	{
 	case ENDF :
-	    if (line == (Token *) 0)
+	    if (line == noToken)
 	    {
 		line = hd;
 	    }
@@ -1143,14 +1596,18 @@ Token *accumulate(Input *file, int nl)
 	    }
 	    file->lineStart = 1;
 	    return line;
+
 	case SEMI:
 	case NL :
-	    if (!nl) { freeToken(hd) ; continue; }
-	    file->lineStart = 1;
-	    goto seminl;
+	    if (!nl)
+	    {
+		freeToken(hd);
+		continue;
+	    }
 	case SCOMMENT :
 	    file->lineStart = 1;
 	    goto seminl;
+
 	case CONST :
 	    if (last == NL && hd->text[0] == '#')
 	    {
@@ -1160,6 +1617,7 @@ Token *accumulate(Input *file, int nl)
 		hd = getComment(file, tb, '\n');
 	    }
 	    break;
+
 	case CONC :
 	    if (last == NL && hd->sequence->type == CPART
 		&& hd->sequence->text[0] == '#')
@@ -1170,9 +1628,11 @@ Token *accumulate(Input *file, int nl)
 		hd = getComment(file, tb, '\n');
 	    }
 	    break;
+
 	case SP :
 	    freeToken(hd);
 	    continue;
+
 	default :
 	    break;
 	}
@@ -1183,3 +1643,154 @@ seminl:
     }
 }
 
+Token *tokacc(Token *tp, int flag, int nl)
+{
+    Input *file = tokenise(tp, flag);
+    Token *lp = accumulate(file, nl);
+
+    untokenise(file);
+    return lp;
+}
+
+static int chkVarToken(Token *token)
+{
+    return (token != noToken &&
+	(token->type == VAR || token->type == CALL || token->type == CONC
+	 || token->type == STRING));
+}
+
+int isVarToken(Token *token)
+{
+    char *cp;
+    TBuff *tb;
+
+    if (token != noToken && token->type == CONC)
+    { /* It might be an array element name */
+        tb = newTB(1024);
+	makeTB(token->sequence, tb);
+	if (tb->buff[tb->length-1] == ')')
+	{
+	    cp = tb->buff;
+	    while (*cp)
+	    {
+	        switch (*cp)
+		{
+		case '[' :
+		case '$' :
+		case ' ' :
+		    freeTB(tb);
+		    return 1;
+		case '(' :
+		    freeTB(tb);
+		    return 0;
+		default:
+		    break;
+		}
+	        cp++;
+	    }
+	}
+	freeTB(tb);
+    }
+    return chkVarToken(token);
+}
+
+int isSingleCall(Token *token, char *proc)
+{
+    if (token != noToken)
+    {
+        switch (token->type)
+	{
+	case CONC :
+	    if (token->sequence->type != CALL || token->sequence->next != (Token *)0)
+	    {
+	        break;
+	    }
+	    token = token->sequence;  /*!!!! Switch drops through here !!!!*/
+	case CALL:
+	    return tokEqual(token->sequence, proc);
+	default :
+	    break;
+	}
+    }
+    return 0;
+}
+
+int isSwitch(Token *token)
+{
+    if (token != noToken)
+    {
+        switch (token->type)
+	{
+	case CONC :
+	    token = token->sequence;
+	    if (token == noToken || token->type != CONST)
+	    {
+	        break;
+	    }
+	/* ********** DROP THROUGH ************** */    
+	case LIST :
+	case CONST:
+	    return tokNEqual(token, "-", 1);
+
+	default :
+	    break;
+	}
+    }
+    return 0;
+}
+
+Token *isArray(Token *tp)
+{
+    Token *res = noToken;
+    TBuff *tbp = (TBuff *) 0;
+    char *txt, *cp, ch;
+    int len;
+
+    if (tp != noToken && (tp->text != (char *) 0 || tp->type == CONC))
+    {
+	switch (tp->type)
+	{
+	case CONC:
+	    tbp = newTB(64);
+	    makeTB(tp->sequence, tbp);
+	    txt = tbp->buff;
+	    len = tbp->length;
+	    break;
+
+	case CONST:
+	case LIST:
+	    txt = tp->text;
+	    len = tp->length;
+	    break;
+
+	default:
+	fprintf(stderr, "default\n");
+	    return noToken;
+	}
+	if (txt[len-1] == ')' && *txt != '(')
+	{ /* could be an array reference.... */
+	    cp = txt;
+	
+	    for(;;)
+	    {
+	        ch = *cp;
+	        if (ch == '(')
+		{
+		    *cp = '\0';
+		    res = createToken(CONST, txt, noToken);
+		    *cp = '(';
+		    break;
+		}
+		if (ch == '\0') { break; }
+		cp++;
+	    }
+	}
+	if (tbp != (TBuff *) 0) { freeTB(tbp); }
+    }
+    return res;
+}
+
+int constant(Token *tp)
+{
+    return (tp != noToken && (tp->type == CONST || tp->type == LIST));
+}
